@@ -12,17 +12,17 @@ environment variable to 2 and then run your code as usual:
   DBI_PROFILE=2 prog.pl
 
 This will profile your program and then output a textual summary
-grouped by query.  You can also enable profiling by setting the
-Profile attribute of any DBI handle:
+grouped by query when the program exits.  You can also enable profiling by
+setting the Profile attribute of any DBI handle:
 
   $dbh->{Profile} = 2;
 
-Other values are possible - see L<"ENABLING A PROFILE"> below.
+Then the summary will be printed when the handle is destroyed.
+
+Many other values apart from are possible - see L<"ENABLING A PROFILE"> below.
 
 =head1 DESCRIPTION
 
-DBI::Profile is new and experimental and subject to change.
-
 The DBI::Profile module provides a simple interface to collect and
 report performance and benchmarking data from the DBI.
 
@@ -52,14 +52,13 @@ calling the appropriate method and just before returning, it takes
 another high-resolution timestamp and calls a function to record
 the information.  That function is passed the two timestamps
 plus the DBI handle and the name of the method that was called.
-That information about a single DBI method call is called the
-I<profile sample> data.
+That data about a single DBI method call is called a I<profile sample>.
 
 =item Data Filtering
 
-If the method call was invoked by the DBI or by a driver then the
-call is currently ignored for profiling because the time spent will
-be accounted for by the original 'outermost' call.
+If the method call was invoked by the DBI or by a driver then the call is
+ignored for profiling because the time spent will be accounted for by the
+original 'outermost' call for your code.
 
 For example, the calls that the selectrow_arrayref() method makes
 to prepare() and execute() etc. are not counted individually
@@ -68,14 +67,14 @@ to the selectrow_arrayref() method when it returns. If this was not
 done then it would be very easy to double count time spent inside
 the DBI.
 
-In future releases it may be possible to alter this behaviour.
-
 =item Data Storage Tree
 
-The profile data is stored as 'leaves on a tree'. The 'path' through
-the branches of the tree to the particular leaf that will store the
-profile sample data for a profiled call is determined dynamically.
-This is a powerful feature.
+The profile data is accumulated as 'leaves on a tree'. The 'path' through the
+branches of the tree to a particular leaf is determined dynamically for each sample.
+This is a key feature of DBI profiliing.
+
+For each profiled method call the DBI walks along the Path and uses each value
+in the Path to step into and grow the Data tree.
 
 For example, if the Path is
 
@@ -85,54 +84,43 @@ then the new profile sample data will be I<merged> into the tree at
 
   $h->{Profile}->{Data}->{foo}->{bar}->{baz}
 
-It wouldn't be very useful to merge all the call data into one leaf
-node (except to get an overall 'time spent inside the DBI' total).
-It's more common to want the Path to include the current statement
-text and/or the name of the method called to show what the time
-spent inside the DBI was for.
+But it's not very useful to merge all the call data into one leaf node (except
+to get an overall 'time spent inside the DBI' total).  It's more common to want
+the Path to include dynamic values such as the current statement text and/or
+the name of the method called to show what the time spent inside the DBI was for.
 
-The Path can contain some 'magic cookie' values that are automatically
-replaced by corresponding dynamic values when they're used.
-For example DBIprofile_Statement (exported by DBI::profile) is
-automatically replaced by value of the C<Statement> attribute of
-the handle. For example, is the Path was:
+The Path can contain some 'magic cookie' values that are automatically replaced
+by corresponding dynamic values when they're used. These magic cookies always
+start with a punctuation character.
 
-  [ 'foo', DBIprofile_Statement, 'bar' ]
+For example a value of 'C<!MethodName>' in the Path causes the corresponding
+entry in the Data to be the name of the method that was called.
+For example, if the Path was:
 
-and the value of $h->{Statement} was:
+  [ 'foo', '!MethodName', 'bar' ]
 
-  SELECT * FROM tablename
+and the selectall_arrayref() method was called, then the profile sample data
+for that call will be merged into the tree at:
 
-then the profile data will be merged into the tree at:
-
-  $h->{Profile}->{Data}->{foo}->{SELECT * FROM tablename}->{bar}
-
-The default Path is just C<[ DBIprofile_Statement ]> and so by
-default the profile data is aggregated per distinct Statement string.
-
-For statement handles this is always simply the string that was
-given to prepare() when the handle was created.  For database handles
-this is the statement that was last prepared or executed on that
-database handle. That can lead to a little 'fuzzyness' because, for
-example, calls to the quote() method to build a new statement will
-typically be associated with the previous statement. In practice
-this isn't a significant issue and the dynamic Path mechanism can
-be used to setup your own rules.
+  $h->{Profile}->{Data}->{foo}->{selectall_arrayref}->{bar}
 
 =item Profile Data
 
 Profile data is stored at the 'leaves' of the tree as references
 to an array of numeric values. For example:
 
-    [
-      106,                    # count
-      0.0312958955764771,     # total duration
-      0.000490069389343262,   # first duration
-      0.000176072120666504,   # shortest duration
-      0.00140702724456787,    # longest duration
-      1023115819.83019,       # time of first event
-      1023115819.86576,       # time of last event
-    ]
+  [
+    106,                  # 0: count of samples at this node
+    0.0312958955764771,   # 1: total duration
+    0.000490069389343262, # 2: first duration
+    0.000176072120666504, # 3: shortest duration
+    0.00140702724456787,  # 4: longest duration
+    1023115819.83019,     # 5: time of first sample
+    1023115819.86576,     # 6: time of last sample
+  ]
+
+After the first sample, later samples always update elements 0, 1, and 6, and
+may update 3 or 4 depending on the duration of the sampled call.
 
 =back
 
@@ -145,52 +133,61 @@ attribute. For example:
 
 The Profile attribute holds a blessed reference to a hash object
 that contains the profile data and attributes relating to it.
+
 The class the Profile object is blessed into is expected to
 provide at least a DESTROY method which will dump the profile data
 to the DBI trace file handle (STDERR by default).
 
-All these examples have the same effect as the first:
+All these examples have the same effect as each other:
 
+  $h->{Profile} = 0;
+  $h->{Profile} = "/DBI::Profile";
+  $h->{Profile} = DBI::Profile->new();
   $h->{Profile} = {};
-  $h->{Profile} = "DBI::Profile";
-  $h->{Profile} = "2/DBI::Profile";
-  $h->{Profile} = 2;
+  $h->{Profile} = { Path => [] };
+
+Similarly, these examples have the same effect as each other:
+
+  $h->{Profile} = 6;
+  $h->{Profile} = "6/DBI::Profile";
+  $h->{Profile} = "!Statement:!MethodName/DBI::Profile";
+  $h->{Profile} = { Path => [ '!Statement', '!MethodName' ] };
 
 If a non-blessed hash reference is given then the DBI::Profile
 module is automatically C<require>'d and the reference is blessed
 into that class.
 
-If a string is given then it is split on 'C</>' characters and the
-first value is used to select the Path to be used (see below).
+If a string is given then it is processed like this:
+
+    ($path, $module, $args) = split /\//, $string, 3
+
+    @path = split /:/, $path
+    @args = split /:/, $args
+
+    eval "require $module" if $module
+    $module ||= "DBI::Profile"
+
+    $module->new( Path => \@Path, @args )
+
+So the first value is used to select the Path to be used (see below).
 The second value, if present, is used as the name of a module which
 will be loaded and it's C<new> method called. If not present it
 defaults to DBI::Profile. Any other values are passed as arguments
-to the C<new> method. For example: "C<2/DBIx::OtherProfile/Foo/42>".
+to the C<new> method. For example: "C<2/DBIx::OtherProfile/Foo:42>".
 
-Various common sequences for Path can be selected by simply assigning
-an integer value to Profile. The simplest way to explain how the
-values are interpreted is to show the code:
+Numbers can be used as a shorthand way to enable common Path values.
+The simplest way to explain how the values are interpreted is to show the code:
 
-    push @Path, "DBI"                       if $path & 0x01;
-    push @Path, DBIprofile_Statement        if $path & 0x02;
-    push @Path, DBIprofile_MethodName       if $path & 0x04;
-    push @Path, DBIprofile_MethodClass      if $path & 0x08;
+    push @Path, "DBI"           if $path_elem & 0x01;
+    push @Path, "!Statement"    if $path_elem & 0x02;
+    push @Path, "!MethodName"   if $path_elem & 0x04;
+    push @Path, "!MethodClass"  if $path_elem & 0x08;
+    push @Path, "!Caller2"      if $path_elem & 0x10;
 
-So using the value "C<1>" causes all profile data to be merged into
-a single leaf of the tree. That's useful when you just want a total.
-
-Using "C<2>" causes profile sample data to be merged grouped by
-the corresponding Statement text. This is the most frequently used.
-
-Using "C<4>" causes profile sample data to be merged grouped by
-the method name ('FETCH', 'prepare' etc.). Using "C<8>" is similar
-but gives the fully qualified 'glob name' of the method called. For
-example: '*DBD::Driver::db::prepare', '*DBD::_::st::fetchrow_hashref'.
-
-The values can be added together to create deeper paths. The most
-useful being 6 (statement then method name) or 10 (statement then
-method name with class).  Using a negative number will reverse the
-path. Thus -6 will group by method name then statement.
+So "2" is the same as "!Statement" and "6" (2+4) is the same as
+"!Statement:!Method".  Those are the two most commonly used values.  Using a
+negative number will reverse the path. Thus "-6" will group by method name then
+statement.
 
 The spliting and parsing of string values assigned to the Profile
 attribute may seem a little odd, but there's a good reason for it.
@@ -198,7 +195,8 @@ Remember that attributes can be embedded in the Data Source Name
 string which can be passed in to a script as a parameter. For
 example:
 
-    dbi:DriverName(RaiseError=>1,Profile=>2):dbname
+    dbi:DriverName(Profile=>2):dbname
+    dbi:DriverName(Profile=>{Username}:!Statement/MyProfiler/Foo:42):dbname
 
 And also, if the C<DBI_PROFILE> environment variable is set then
 The DBI arranges for every driver handle to share the same profile
@@ -219,49 +217,38 @@ A reference to a hash containing the collected profile data.
 
 =head2 Path
 
-The Path value is used to control where the profile for a method
-call will be merged into the collected profile data.  Whenever
-profile data is to be stored the current value for Path is used.
+The Path value is a reference to an array. Each element controls the
+value to use at the corresponding level of the profile Data tree.
 
-The value can be one of:
+The elements of Path array can be one of the following types:
 
 =over 4
 
-=item Array Reference
+=item Special Constant
 
-Each element of the array defines an element of the path to use to
-store the profile data into the C<Data> hash.
+B<!Statement>
 
-=item Undefined value (the default)
+Use the current Statement text. Typically that's the value of the Statement
+attribute for the handle the method was called with. Some methods, like
+commit() and rollback(), are unrelated to a particular statement. For those
+methods !Statement records an empty string.
 
-Treated the same as C<[ $DBI::Profile::DBIprofile_Statement ]>.
+For statement handles this is always simply the string that was
+given to prepare() when the handle was created.  For database handles
+this is the statement that was last prepared or executed on that
+database handle. That can lead to a little 'fuzzyness' because, for
+example, calls to the quote() method to build a new statement will
+typically be associated with the previous statement. In practice
+this isn't a significant issue and the dynamic Path mechanism can
+be used to setup your own rules.
 
-=item Subroutine Reference B<NOT YET IMPLEMENTED>
+B<!MethodName>
 
-The subroutine is passed the DBI method name and the handle it was
-called on.  It should return a list of values to uses as the path.
-If it returns an empty list then the method call is not profiled.
+Use the name of the DBI method that the profile sample relates to.
 
-=back
+B<!MethodClass>
 
-The following 'magic cookie' values can be included in the Path and will be
-
-=over 4
-
-=item DBIprofile_Statement
-
-Replaced with the current value of the Statement attribute for the
-handle the method was called with. If that value is undefined then
-an empty string is used.
-
-=item DBIprofile_MethodName
-
-Replaced with the name of the DBI method that the profile sample
-relates to.
-
-=item DBIprofile_MethodClass
-
-Replaced with the fully qualified name of the DBI method, including
+Use the fully qualified name of the DBI method, including
 the package, that the profile sample relates to. This shows you
 where the method was implemented. For example:
 
@@ -275,22 +262,84 @@ inherited the selectrow_arrayref method provided by the DBI.
 
 But you'll note that there is only one call to
 DBD::_::db::selectrow_arrayref but another 99 to
-DBD::mysql::db::selectrow_arrayref. That's because after the first
-call Perl has cached the method to speed up method calls.
-You may also see some names begin with an asterix ('C<*>').
-Both of these effects are subject to change in later releases.
+DBD::mysql::db::selectrow_arrayref. Currently the first
+call Pern't record the true location. That may change.
 
+B<!Caller>
+
+Use a string showing the filename and line number of the code calling the method.
+
+B<!Caller2>
+
+Use a string showing the filename and line number of the code calling the
+method, as for !Caller, but also include filename and line number of the code
+that called that. Calls from DBI:: and DBD:: packages are skipped.
+
+B<!File>
+
+Same as !Caller above except that only the filename is included, not the line number.
+
+B<!File2>
+
+Same as !Caller2 above except that only the filenames are included, not the line number.
+
+=item Code Reference
+
+The subroutine is passed the handle it was called on and the DBI method name.
+The current Statement is in $_. The statement string should not be modified,
+so most subs start with C<local $_ = $_;>.
+
+The list of values it returns is used at that point in the Profile Path.
+
+The sub can 'veto' (reject) a profile sample by including a reference to undef
+in the returned list. That can be useful when you want to only profile
+statements that match a certain pattern, or only profile certain methods.
+
+=item Subroutine Specifier
+
+A Path element that begins with 'C<&>' is treated as the name of a subroutine
+in the DBI::ProfileSubs namespace and replaced with the corresponding code reference.
+
+Currently this only works when the Path is specified by the C<DBI_PROFILE>
+environment variable.
+
+Also, currently, the only subroutine in the DBI::ProfileSubs namespace is
+C<'&norm_std_n3'>. That's a very handy subroutine when profiling code that
+doesn't use placeholders. See L<DBI::ProfileSubs> for more information.
+
+=item Attribute Specifier
+
+A string enclosed in braces, such as 'C<{Username}>', specifies that the current
+value of the corresponding database handle attribute should be used at that
+point in the Path.
+
+=item Reference to a Scalar
+
+Specifies that the current value of the referenced scalar be used at that point
+in the Path.  This provides an efficient way to get 'contextual' values into
+your profile.
+
+=item Other Values
+
+Any other values are stringified and used literally.
+
+(References, and values that begin with punctuation characters are reserved.)
 
 =back
 
-Other magic cookie values may be added in the future.
+Only the first 100 elements in Path are used.
+
+If the value of Path is anything other than an array reference,
+it is treated as if it was:
+
+	[ DBI::Profile::!Statement ]
 
 
 =head1 REPORTING
 
 =head2 Report Format
 
-The current profile data can be formatted and output using
+The current accumulated profile data can be formatted and output using
 
     print $h->{Profile}->format;
 
@@ -302,7 +351,7 @@ you can do:
 
 The default results format looks like this:
 
-  DBI::Profile: 0.001015 seconds (5 method calls) programname
+  DBI::Profile: 0.001015s 42.7% (5 calls) programname @ YYYY-MM-DD HH:MM:SS
   '' =>
       0.000024s / 2 = 0.000012s avg (first 0.000015s, min 0.000009s, max 0.000015s)
   'SELECT mode,size,name FROM table' =>
@@ -315,7 +364,8 @@ run, then a formated version of the profile data tree.
 If the results are being formated when the perl process is exiting
 (which is usually the case when the DBI_PROFILE environment variable
 is used) then the percentage of time the process spent inside the
-DBI is also shown.
+DBI is also shown. If the process is not exiting then the percentage is
+calculated using the time between the first and last call to the DBI.
 
 In the example above the paths in the tree are only one level deep and
 use the Statement text as the value (that's the default behaviour).
@@ -352,7 +402,7 @@ time per leaf node.
 The default method of reporting is for the DESTROY method of the
 Profile object to format the results and write them using:
 
-    DBI->trace_msg($results, 0)
+    DBI->trace_msg($results, 0);  # see $ON_DESTROY_DUMP below
 
 to write them to the DBI trace() filehandle (which defaults to
 STDERR). To direct the DBI trace filehandle to write to a file
@@ -364,6 +414,14 @@ trace level of 0. For example:
 The same effect can be achieved without changing the code by
 setting the C<DBI_TRACE> environment variable to C<0=filename>.
 
+The $DBI::Profile::ON_DESTROY_DUMP variable holds a code ref
+that's called to perform the output of the formatted results.
+The default value is:
+
+  $ON_DESTROY_DUMP = sub { DBI->trace_msg($results, 0) };
+
+Apart from making it easy to send the dump elsewhere, it can also
+be useful as a simple way to disable dumping results.
 
 =head1 CHILD HANDLES
 
@@ -373,6 +431,52 @@ then by default the statement handles created from it all contribute
 to the same merged profile data tree.
 
 
+=head1 CUSTOM DATA MANIPULATION
+
+Recall that C<$h->{Profile}->{Data}> is a reference to the collected data.
+Either to a 'leaf' array (when the Path is empty, i.e., DBI_PROFILE env var is 1),
+or a reference to hash containing values that are either further hash
+references or leaf array references.
+
+Sometimes it's useful to be able to summarise some or all of the collected data.
+The dbi_profile_merge() function can be used to merge leaf node values.
+
+=head2 dbi_profile_merge
+
+  use DBI qw(dbi_profile_merge);
+
+  $time_in_dbi = dbi_profile_merge(my $totals=[], @$leaves);
+
+Merges profile data node. Given a reference to a destination array, and zero or
+more references to profile data, merges the profile data into the destination array.
+For example:
+
+  $time_in_dbi = dbi_profile_merge(
+      my $totals=[],
+      [ 10, 0.51, 0.11, 0.01, 0.22, 1023110000, 1023110010 ],
+      [ 15, 0.42, 0.12, 0.02, 0.23, 1023110005, 1023110009 ],
+  );        
+
+$totals will then contain
+
+  [ 25, 0.93, 0.11, 0.01, 0.23, 1023110000, 1023110010 ]
+
+and $time_in_dbi will be 0.93;
+
+For example, to get the time spent 'inside' the DBI during an http request,
+your logging code run at the end of the request (i.e. mod_perl LogHandler)
+could use:
+
+  my $time_in_dbi = 0;
+  if (my $Profile = $dbh->{Profile}) { # if DBI profiling is enabled
+      $time_in_dbi = dbi_profile_merge(my $total=[], $Profile->{Data});
+      $Profile->{Data} = {}; # reset the profile data
+  }
+
+If profiling has been enabled then $time_in_dbi will hold the time spent inside
+the DBI for that handle (and any other handles that share the same profile data)
+since the last request.
+
 =head1 CUSTOM DATA COLLECTION
 
 =head2 Using The Path Attribute
@@ -392,7 +496,7 @@ into the profile data tree. For example:
     use DBI;
     use DBI::Profile (dbi_profile dbi_time);
 
-    my $t1 = dbi_time(); # returns floatingpoint high-resolution time
+    my $t1 = dbi_time(); # floating point high-resolution time
 
     ... execute code you want to profile here ...
 
@@ -401,9 +505,9 @@ into the profile data tree. For example:
 
 The $h parameter is the handle the extra profile sample should be
 associated with. The $statement parameter is the string to use where
-the Path specifies DBIprofile_Statement. If $statement is undef
+the Path specifies !Statement. If $statement is undef
 then $h->{Statement} will be used. Similarly $method is the string
-to use if the Path specifies DBIprofile_MethodName. There is no
+to use if the Path specifies !MethodName. There is no
 default value for $method.
 
 The $h->{Profile}{Path} attribute is processed by dbi_profile() in
@@ -425,8 +529,9 @@ they work with future versions of the DBI.
 
 Applications which generate many different statement strings
 (typically because they don't use placeholders) and profile with
-DBIprofile_Statement in the Path (the default) will consume memory
-in the Profile Data structure for each statement.
+!Statement in the Path (the default) will consume memory
+in the Profile Data structure for each statement. Use a code ref
+in the Path to return an edited (simplified) form of the statement.
 
 If a method throws an exception itself (not via RaiseError) then
 it won't be counted in the profile.
@@ -443,12 +548,22 @@ DBI->connect is counted if the DBI_PROFILE environment variable is set.
 
 Time spent fetching tied variables, $DBI::errstr, is counted.
 
+Time spent in FETCH for $h->{Profile} is not counted, so getting the profile
+data doesn't alter it.
+
 DBI::PurePerl does not support profiling (though it could in theory).
 
 A few platforms don't support the gettimeofday() high resolution
 time function used by the DBI (and available via the dbi_time() function).
 In which case you'll get integer resolution time which is mostly useless.
 
+On Windows platforms the dbi_time() function is limited to millisecond
+resolution. Which isn't sufficiently fine for our needs, but still
+much better than integer resolution. This limited resolution means
+that fast method calls will often register as taking 0 time. And
+timings in general will have much more 'jitter' depending on where
+within the 'current millisecond' the start and and timing was taken.
+
 This documentation could be more clear. Probably needs to be reordered
 to start with several examples and build from there.  Trying to
 explain the concepts first seems painful and to lead to just as
@@ -458,14 +573,15 @@ many forward references.  (Patches welcome!)
 
 
 use strict;
-use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION);
+use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION $ON_DESTROY_DUMP);
 use Exporter ();
 use UNIVERSAL ();
 use Carp;
 
 use DBI qw(dbi_time dbi_profile dbi_profile_merge);
 
-$VERSION = sprintf "%d.%02d", '$Revision: 1.5 $ ' =~ /(\d+)\.(\d+)/;
+$VERSION = sprintf("2.%06d", q$Revision: 9294 $ =~ /(\d+)/o);
+
 
 @ISA = qw(Exporter);
 @EXPORT = qw(
@@ -480,10 +596,11 @@ $VERSION = sprintf "%d.%02d", '$Revision: 1.5 $ ' =~ /(\d+)\.(\d+)/;
     format_profile_thingy
 );
 
-use constant DBIprofile_Statement	=> -2100000001;
-use constant DBIprofile_MethodName	=> -2100000002;
-use constant DBIprofile_MethodClass	=> -2100000003;
+use constant DBIprofile_Statement	=> '!Statement';
+use constant DBIprofile_MethodName	=> '!MethodName';
+use constant DBIprofile_MethodClass	=> '!MethodClass';
 
+$ON_DESTROY_DUMP = sub { DBI->trace_msg(shift, 0) };
 
 sub new {
     my $class = shift;
@@ -498,46 +615,50 @@ sub _auto_new {
 
     # This sub is called by DBI internals when a non-hash-ref is
     # assigned to the Profile attribute. For example
-    #	dbi:mysql(RaiseError=>1,Profile=>4/DBIx::MyProfile):dbname
+    #	dbi:mysql(RaiseError=>1,Profile=>!Statement:!MethodName/DBIx::MyProfile/arg1:arg2):dbname
     # This sub works out what to do and returns a suitable hash ref.
     
-    my ($path, $module, @args);
-
-    # parse args
-    if ($arg =~ m!/!) {
-        # it's a path/module/arg/arg/arg list
-        ($path, $module, @args) = split /\s*\/\s*/, $arg, -1;
-    } elsif ($arg =~ /^\d+$/) {
-        # it's a numeric path selector
-        $path = $arg;
-    } else {
-        # it's a module name
-        $module = $arg;
-    }
+    $arg =~ s/^DBI::/2\/DBI::/
+        and carp "Automatically changed old-style DBI::Profile specification to $arg";
 
+    # it's a path/module/arg/arg/arg list
+    my ($path, $package, $args) = split /\//, $arg, 3;
+    my @args = (defined $args) ? split(/:/, $args, -1) : ();
     my @Path;
-    if ($path) {
-	my $reverse = ($path < 0) ? ($path=-$path, 1) : 0;
-	push @Path, "DBI"			if $path & 0x01;
-	push @Path, DBIprofile_Statement	if $path & 0x02;
-	push @Path, DBIprofile_MethodName	if $path & 0x04;
-	push @Path, DBIprofile_MethodClass	if $path & 0x08;
-	@Path = reverse @Path if $reverse;
-    } else {
-        # default Path
-        push @Path, DBIprofile_Statement;
+
+    for my $element (split /:/, $path) {
+        if (DBI::looks_like_number($element)) {
+            my $reverse = ($element < 0) ? ($element=-$element, 1) : 0;
+            my @p;
+            # a single "DBI" is special-cased in format()
+            push @p, "DBI"			if $element & 0x01;
+            push @p, DBIprofile_Statement	if $element & 0x02;
+            push @p, DBIprofile_MethodName	if $element & 0x04;
+            push @p, DBIprofile_MethodClass	if $element & 0x08;
+            push @p, '!Caller2'            	if $element & 0x10;
+            push @Path, ($reverse ? reverse @p : @p);
+        }
+        elsif ($element =~ m/^&(\w.*)/) {
+            my $name = "DBI::ProfileSubs::$1"; # capture $1 early
+            require DBI::ProfileSubs;
+            my $code = do { no strict; *{$name}{CODE} };
+            if (defined $code) {
+                push @Path, $code;
+            }
+            else {
+                warn "$name: subroutine not found\n";
+                push @Path, $element;
+            }
+        }
+        else {
+            push @Path, $element;
+        }
     }
 
-    if ($module) {
-	if (eval "require $module") {
-	  $class = $module;
-	}
-	else {
-	    carp "Can't use $module for DBI profile: $@";
-	}
-    }
+    eval "require $package" if $package; # sliently ignores errors
+    $package ||= $class;
 
-    return $class->new(Path => \@Path, @args);
+    return $package->new(Path => \@Path, @args);
 }
 
 
@@ -554,18 +675,19 @@ sub format {
 
     if (@$leaves) {
 	dbi_profile_merge(my $totals=[], @$leaves);
-	my ($count, $dbi_time) = @$totals;
+	my ($count, $time_in_dbi, undef, undef, undef, $t1, $t2) = @$totals;
 	(my $progname = $0) =~ s:.*/::;
 	if ($count) {
-	    $prologue .= sprintf "%f seconds ", $dbi_time;
-	    my $perl_time = dbi_time() - $^T;
-	    $prologue .= sprintf "%.2f%% ", $dbi_time/$perl_time*100
-		if $DBI::PERL_ENDING && $perl_time;
-	    $prologue .= sprintf "(%d method calls) $progname\n", $count;
+	    $prologue .= sprintf "%fs ", $time_in_dbi;
+	    my $perl_time = ($DBI::PERL_ENDING) ? time() - $^T : $t2-$t1;
+	    $prologue .= sprintf "%.2f%% ", $time_in_dbi/$perl_time*100 if $perl_time;
+	    my @lt = localtime(time);
+	    my $ts = sprintf "%d-%02d-%02d %02d:%02d:%02d",
+		1900+$lt[5], $lt[4]+1, @lt[3,2,1,0];
+	    $prologue .= sprintf "(%d calls) $progname \@ $ts\n", $count;
 	}
-
-	if (@$leaves == 1 && $self->{Data}->{DBI}) {
-	    $detail = "";	# hide it
+	if (@$leaves == 1 && ref($self->{Data}) eq 'HASH' && $self->{Data}->{DBI}) {
+	    $detail = "";	# hide the "DBI" from DBI_PROFILE=1
 	}
     }
     return ($prologue, $detail) if wantarray;
@@ -613,6 +735,7 @@ sub format_profile_branch {
 
 sub format_profile_thingy {
     my ($self, $thingy, $depth, $pad, $path, $leaves) = @_;
+    return "undef" if not defined $thingy;
     return $self->format_profile_leaf(  $thingy, $depth, $pad, $path, $leaves)
 	if UNIVERSAL::isa($thingy,'ARRAY');
     return $self->format_profile_branch($thingy, $depth, $pad, $path, $leaves)
@@ -623,12 +746,15 @@ sub format_profile_thingy {
 
 sub on_destroy {
     my $self = shift;
-    my $detail = $self->format() if $self->{Data};
-    DBI->trace_msg($detail, 0) if $detail;
+    return unless $ON_DESTROY_DUMP;
+    return unless $self->{Data};
+    my $detail = $self->format();
+    $ON_DESTROY_DUMP->($detail) if $detail;
 }
 
 sub DESTROY {
     my $self = shift;
+    local $@;
     eval { $self->on_destroy };
     if ($@) {
         my $class = ref($self) || $self;
